#/usr/bin/env python
#!/usr/bin/env python
#coding:utf-8
# Author:   Yaron Butterfield
# Purpose: For RNASeq-Align.  If this cigar (after alignment to genome + junctions) 
#          contains S, I, or D tags, convert the cigar to include junction regions
#    
# Created: 7/20/2011 / original 2009/2010

import sys
   


def read_cigar(cigar):
    cigar_format=["M","I","D","N","S","H","P"]
    coord=''
    c=[]
    for i in cigar:
	if i in cigar_format:
	    code = coord+i
	    c.append(code)
	    coord=''
	else:
	    coord+=i
    return c
   

def cigarValues(c):
    c1=[]
    cN=[]
    cM=[]
    cS=[]
    for k in c:
	c1.append([int(k[:-1]),k[-1]])
	if k[-1]=='N':
	    cN.append(k[:-1])
	elif k[-1]=='M':
	    cM.append(int(k[:-1]))
	elif k[-1]=='S':
	    cM.append(int(k[:-1]))
    return c1, cN, cM, cS


def getCigarValues(cigar):
    
    c = read_cigar(cigar)
    
    x = cigarValues(c)
    totalCov=0
    for cov in x[0]:
	totalCov += int(cov[0])
	
    return totalCov



def convert(cigar,zcigar,seqLength):
    # note that the intial alignment can have softclips.
    # i.e. zcigar can have softclips
    #print "Converting:",cigar,zcigar
    nC=''
    
    c = read_cigar(cigar) #  
    j = read_cigar(zcigar) # <-- original cigar i.e.  2M3I45M
    c1=[]
    cN=[]
    cM=[]
    cS=[]
    for k in c:
	c1.append([int(k[:-1]),k[-1]])
	if k[-1]=='N':
	    cN.append(k[:-1])
	elif k[-1]=='M':
	    cM.append(int(k[:-1]))
	elif k[-1]=='S':
	    cM.append(int(k[:-1]))
		
    j1=[]
    jID=[]
    jM=[]
    MTotal=seqLength
    
    for k in j:
	j1.append([int(k[:-1]),k[-1]])
	
	if k[-1]=='I' or k[-1]=='D':
	    jID.append(k)
	elif k[-1]=='M':
	    jM.append(int(k[:-1]))
	    
	
    k=0
    diff=0
    for i in cM:
	sc=0  # start softclip amount
	sc2=0 # end softclip amount
	
	if c1[0][1]=='S':
	    sc = c1[0][0] # start softclip amount
	    #nC += str(c1[0][0]) + "S"
	    #MTotal-=c1[0][0]
	    
	if c1[-1][1]=='S':
	    sc2 = c1[-1][0] # end softclip amount
	    
	if k==0:
	    delta = i - j1[0][0]
	    h = j1[0][0]
	    #print i, delta,h
	    
	    if delta > 0 and sc ==0:
		
		#print "?"
		if j1[0][1] == "M":
		    nC += str(j1[0][0]) + "M"
		    MTotal-=j1[0][0]
		    
		    j1.pop(0)
		elif j1[0][1] == "S":
		    nC += str(j1[0][0]) + "S"
		    MTotal-=j1[0][0]
		    j1.pop(0)
		
		#print nC,j1
		
		# most likely an indel cigar
		nC += str(j1[0][0]) + j1[0][1]
		#print nC
		indel=0
		if j1[0][1] == 'D':
		    indel = j1[0][0]
		    
		if j1[0][1] == "I":
		    MTotal-=j1[0][0]
		j1.pop(0)
		    
		if indel + h <= i:
		    #print indel, h
		    delta -= indel
		    nC += str(delta) + "M"
		    MTotal-=delta
		    
		#print nC
		if len(cN)>0:
		    nC += cN.pop(0)+"N"
		    
	    elif delta > 0 and sc > 0:
		# situation where start of read aligns to previous junction (over the N's)
		nC += c[0]
		MTotal-= i
		
		#------------
		#delta-=sc
		
		#j1.pop(0)
		
		#if j1[0][1] == "M":
		    #nC += str(j1[0][0]) + "M"
		    #MTotal-=j1[0][0]
		    
		    #j1.pop(0)
		#-------------------
		    
		if j1[0][1] == "D" and j1[0][0]==1 and len(cN)==0:
		    break
		if j1[0][1] == "D" and j1[0][0]==1 and len(cN)>0:
		    j1.pop(0)
		

	    elif delta < 0 and sc2 ==0:
		nC += c[0]
		#print nC
		MTotal-= i
		j1.pop(0)
		
		if len(cN)>0:
		    nC += cN.pop(0)+"N"
		    
		diff=0
		if len(cM) > 1:
		    diff = delta + cM[1] 

		if j1[0][0]>diff:
		    nC += str(abs(delta)+diff)+"M"
		    MTotal-=(abs(delta)+diff)
		    
		    #nC += str(j1[0][0]) + j1[0][1]
		    #if j1[0][1] == "I":
			#MTotal-=j1[0][0]
		    #j1.pop(0)
		else:
		    nC += str(abs(delta))+"M"
		    MTotal-=abs(delta)
		

		#print nC,MTotal
		
	    elif delta < 0 and sc2 > 0:
		nC += c[0]
		MTotal-= i
		j1.pop(0)
	
		if j1[0][1] == "D" and j1[0][0]==1 and len(cN)==0:
		    break
		if j1[0][1] == "D" and j1[0][0]==1 and len(cN)>0:
		    j1.pop(0)
		if j1[0][1] == "I" and len(cN)==0:
		    remainingM = j1[0][0] + j1[1][0]
		    if remainingM < sc2:
			nC += str(sc2) + "S"
			MTotal-= sc2
			break
		    
	    elif delta==0:
		nC += c[0]
		MTotal-= i
		
		if len(cN)>0:
		    nC += cN.pop(0)+"N"
		j1.pop(0)

	elif k==1:
	    
	    if delta == 0:

		if j1[0][1] == "I" or j1[0][1] == "D":
		    nC += str(j1[0][0]) + j1[0][1]
		    if j1[0][1] == "I":
			MTotal-=j1[0][0]
		    else:
			nC += str(i) + "M"
			MTotal-=i
		    
		if len(cN)>0:
		    nC += cN.pop(0)+"N"
		    
		    
	    if delta > 0 :
		
		#print j1
		
		if len(j1)>0:
		    if j1[0][1] == "I" or j1[0][1] == "D" :
		    
			nC += str(j1[0][0]) + j1[0][1]
			if j1[0][1] == "I":
			    MTotal-=j1[0][0]
		    else:
			if MTotal < str(i) and len(c)<4:
			    nC += str(MTotal) + "M"
			    MTotal-=MTotal
			else:
			    nC += str(i) + "M"
			    MTotal -= i
		else:
		    if MTotal < str(i) and len(c)<4:
			nC += str(MTotal) + "M"
			MTotal-=MTotal
		    else:
			nC += str(i) + "M"
			MTotal -= i
		    
		    
		if len(cN)>0:
		    nC += cN.pop(0)+"N"
		
	    if delta < 0:
		#if diff > 0:
		    #nC += str(diff) + "M"
		    #MTotal-=diff
		    
		if len(cN)>0:
		    nC += cN.pop(0)+"N"
		    
		if j1[0][1] == "I" or j1[0][1] == "D":
		    nC += str(j1[0][0]) + j1[0][1]
		    if j1[0][1] == "I":
			MTotal-=j1[0][0]
		else:
		    #nC += str(i) + "M"
		    nC += str(MTotal) + "M"
		    MTotal-=MTotal

	elif k==2:
	    if MTotal > seqLength:
		nC += str(MTotal) + "M"
		MTotal=0
	    
	k+=1
	
    if MTotal > 0:
	nC += str(MTotal) + "M"
	
    #print "->",nC
    #------------------------------------------------------    
    # POST checks
    # M check: M count and back to back M's.    
    # Length:  Make sure M + I + S == sequence length
    nC = doubleMcheck(nC)    
    #print nC
    
    # double check the cigar length 
    # should be correct by now and if not, do not use, return '*'
    x = cigarCheck(nC,c,seqLength)
    if x == 0:
	nC = '*'
	
    return nC
    

def temp():
        #print "nC2=>",nC2
    modified=0
    ## S check -> assumes only one S as reads with softclips on both ends will be excluded
    #if "S" in cigar:
	#modified=1
	#c = read_cigar(cigar) 
	#print c
	#s = 0  # beginning (start) or end (stop0
	#sc = 0 # soft clip amount
	#for k in c:
	    #if k[-1]=='S':
		#sc = int(k[:-1])
		#break
	    #else:
		#s+=1
	
	#if s == 0:
	    #miniCigar = [sc,'S']
	    #diff2 = sc - nC2[0][0]
	    #nC2.pop(0)
	    #nC2.insert(0,miniCigar)
	    #if diff2>0:
		#nC2.pop(1)
	    ##print nC2
	    ##nC2[1][0]-=sc
	    #print nC2
	    
	#else:
	    #miniCigar = [sc,'S']
	    
	    #nC2.insert(nC2Length,miniCigar)
	    
	    #nC2[-2][0]-=sc
    pass


def cigarCheck(nC1,c,seqLength):
    # double check the cigar length 
    c1=[]
    #print c
    Mcount=0 # number of M, S, or I aligned bases; must equal sequence length
    
    #print "--------------------------------------------||||||||||||||||||||---->",c
    
    nC = read_cigar(nC1)
    
    #print nC[0][-1],nC[1][-1]
    
    if nC[0][-1]=='S' and nC[1][-1]=='N':
	return 0
    
    for k in c:
	if k[-1]=='M':
	    Mcount += int(k[:-1])
	elif k[-1]=='S':
	    Mcount += int(k[:-1])
	elif k[-1]=='I':
	    Mcount += int(k[:-1])
	    
    if Mcount <> seqLength:
	#print "?",Mcount, seqLength
	return 0
    else:
	return 1

    
def doubleMcheck(nC):
    # check for back to back M's
    # how much does this process slow the overall code?
    # shouldn't this be avoided in the earlier code?
    i=0
    Mcount=0    
    nC1 = read_cigar(nC)
    nC2=[]
    for k in nC1:
	nC2.append([int(k[:-1]),k[-1]])
    nC2Length = len(nC2)

    while i < len(nC2)-1:
	#print i
	if nC2[i][1]=="M" and nC2[i][1] == nC2[i+1][1]:
	    newM = nC2[i][0] + nC2[i+1][0]
	    nC2[i][0] = newM
	    if i > 1 or i+2==len(nC2):
		nC2.pop()
		break
	    else:
		nC2.pop(1)
		break
	i+=1
    nC=''
    for i in nC2:
	nC+=str(i[0])+i[1]
	
    return nC

def run_100mer_tests():
    seqlength=100
    
    
    origCigars = ["5M1D95M"]
    tempCigars = ["7M528N83M545N10M"]
    correctCigars = ["5M1D1M528N83M545N11M"]
   
    c=0

    if len(origCigars) <> len(tempCigars) and len(tempCigars) <> len(correctCigars) and len(correctCigars) <> len(origCigars):
	print "Number of CIGARS are not matching"
	exit(0)
	
    totalCorrect = 0
    incorrect=[]
    for i in origCigars:
	
	#print "Example:"
	
	zcigar = origCigars[c]
	cigar = tempCigars[c]
	correct = correctCigars[c]
	
	newcigar = convert(cigar,zcigar,seqlength)
	testNumber = "Test " + str(c+1)
	print "                                 ",testNumber
	print "z.cigar",zcigar
	print "cigar", cigar
	print "new",newcigar
	print "cor",correct
	
	if correct==newcigar:
	    totalCorrect+=1
	    print " :D  Yes!"
	else:
	    print " :(  Oops"
	    incorrect.append(testNumber)
	    
	print '----------------------------------------'
	
	c+=1

	
    print "Total examined:",str(len(origCigars))
    print "Total correct cigars:",str(totalCorrect)
    if len(origCigars) == totalCorrect:
	print "All cigars are correct!!"
    else:
	print "Incorrect cigar(s):", incorrect
	
	
def run_75mer_tests():
    seqlength=75
    # 75 mers 
    
    
    origCigars = ["26M4I45M","4M3I68M","50M1D25M","26M4I45M","8M1I66M",
                  "3M1D72M","5M1D70M","5M1I69M","46M3D29M","6M1D69M",
                  "16M3I56M","68M1I6M","71M1D4M","4S71M","45M3D30M","18M2D57M","5M1D70M"]
    
    tempCigars = ["26M3078N49M","8M1500N25M2340N42M",'4S71M','26M3078N49M',"7M815N68M",
                  "30M1000N45M","6S60M150N9M","6S60M150N9M","3M1677N44M2019N28M","7S68M",
                  "19M2175N56M","67M8S","18M24240N52M5S","73M837N2M","47M3487N28M","1S73M857N1M","7M528N68M"]
    
    correctCigars = ["26M3078N4I45M","4M3I4M1500N25M2340N39M","4S46M1D25M","26M3078N4I45M","7M815N1M1I66M",
                     "3M1D26M1000N46M","6S60M150N9M","6S60M150N9M","3M1677N44M2019N3D28M","7S68M",
                     "16M3I3M2175N53M","67M8S","18M24240N57M","4S69M837N2M","45M3D3487N30M","1S17M2D56M857N1M","5M1D1M528N69M"]
   
    c=0

    if len(origCigars) <> len(tempCigars) and len(tempCigars) <> len(correctCigars) and len(correctCigars) <> len(origCigars):
	print "Number of CIGARS are not matching"
	exit(0)
	
    totalCorrect = 0
    incorrect=[]
    for i in origCigars:
	
	#print "Example:"
	
	zcigar = origCigars[c]
	cigar = tempCigars[c]
	correct = correctCigars[c]
	
	newcigar = convert(cigar,zcigar,seqlength)
	testNumber = "Test " + str(c+1)
	print "                                 ",testNumber
	print "z.cigar",zcigar
	print "cigar", cigar
	print "new",newcigar
	print "cor",correct
	
	if correct==newcigar:
	    totalCorrect+=1
	    print " :D  Yes!"
	else:
	    print " :(  Oops"
	    incorrect.append(testNumber)
	    
	print '----------------------------------------'
	
	c+=1

	
    print "Total examined:",str(len(origCigars))
    print "Total correct cigars:",str(totalCorrect)
    if len(origCigars) == totalCorrect:
	print "All cigars are correct!!"
    else:
	print "Incorrect cigar(s):", incorrect
	
	

def get_ZJ_tag(cigar,pos):
    # custom feature requested by Nina Theissen which lists first and last bases of intron(s) skipped
    # Uses information from repositioned coordinate and cigar
    
    """
     14786,'44M140N6M1000N25M'
    """
    
    pos = int(pos)

    c = read_cigar(cigar)  
    
    #print c
    
    ZJtag=''
    exonAligned=0
    intronSize=0
    pairs=[]
    
    for k in c:

	if k[-1]=='N':
	    intronSize = int(k[:-1])
	     
	    pairs.append(str(pos + exonAligned)+"-"+str(pos + exonAligned + intronSize -1))
	    pos = pos + exonAligned + intronSize 
	    
	    exonAligned =0
	elif k[-1]=='M':
	    exonAligned += int(k[:-1])
	elif k[-1]=='D':
	    exonAligned += int(k[:-1])

	    
    ZJtag = 'ZJ:Z:'+ ','.join(pairs)
    
    return ZJtag	
	

	
def run_ZJ_tests():
    
    tests = [[14786,'44M140N6M'],
             [14786,'44M140N6M1000N25M'],
             [44047613,'5M1D2M528N83M545N10M']
             ]
    
    for i in tests:
	pos = i[0]
	cigar = i[1]
	
	ZJtag = get_ZJ_tag(cigar,pos)	
	print ZJtag
    
	
    
if __name__ == '__main__': 
    run_75mer_tests()
    #run_100mer_tests()
    #run_ZJ_tests()